/*
 * Copyright (c) 1998, 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <assert.h>
#include <mom/libmint.h>
#include <mom/c/libpres_c.h>
#include <mom/c/libcast.h>
#include <mom/c/scml.hh>
#include "iiopxx.h"

/* This returns a presentation_collection referenced by the presentation
   implementation file.  The collection is usually created by some
   specific function that can link together additional functions and what not.
*/
presentation_collection *get_iiopxx_pres_impl(struct scml_context *sc)
{
	presentation_collection *retval = 0;
	struct scml_scope *root_scope;
	struct scml_stream_pos *ssp;
	tag_item *ti;
	
	root_scope = sc->get_scope();
	ssp = sc->get_stream_pos();
	/* Get the name of the presentation_implementation */
	if( (ti = find_tag(sc->get_scope()->get_values(),
			   "presentation_implementation_name")) ) {
		char *impl_type;
		
		if( (impl_type = scml_string::tag_string(ti)) ) {
			/* Match the name to some function */
			if( !strcmp(impl_type, "tao_cxx") )
				retval = make_tao_impl(root_scope, ssp);
			else
				panic("Couldn't match "
				      "presentation_implementation_name '%s'"
				      " to any known implementation",
				      impl_type);
		} else {
			panic("Couldn't get string from "
			      "presentation_implementation_name tag");
		}
	} else {
		panic("Couldn't find presentation_implementation_name "
		      "in root scope");
	}
	return( retval );
}

void init_be(be_flags the_flags)
{
	struct scml_scope *root_scope = scml_scope::make_root_scope();
	struct scml_stream_pos *ssp = 0;
	FILE *file;
	char *found_filename;
	
	cast_language = CAST_CXX;
	/* Read in the presentation implementation
	   file and set everything up */
	file = fopen_search(the_flags.pres_impl,
			    the_flags.pres_impl_dirs,
			    "r",
			    &found_filename);
	if (file) {
		struct scml_token_sequence *sts;
		struct scml_stream *ss;
		struct scml_parser *sp;
		
		ss = new scml_stream;
		ss->set_file(file);
		ss->set_desc(the_flags.pres_impl);
		ss->set_include_directory_list(the_flags.pres_impl_dirs);
		ssp = new scml_stream_pos;
		ssp->set_stream(ss);
		sp = new scml_parser;
		sp->set_stream_pos(ssp);
		if( (sts = sp->parse()) ) {
			struct scml_context *sc;
			
			sc = new scml_context;
			sc->set_flags(sc->get_flags() |
				      SCF_IGNORE_WHITE_SPACE);
			sc->set_stream_pos(ssp);
			sc->set_scope(root_scope);
			if( (sc->format_sequence(sts)) ) {
				pres_coll = get_iiopxx_pres_impl(sc);
			} else {
				panic("There was an error in executing the"
				      " presentation implementation");
			}
		} else {
			panic("Couldn't parse the presentation"
			      "implementation");
		}
		fclose(file);
	} else {
		panic("Unable to open the presentation implementation "
		      "file '%s'", the_flags.pres_impl);
	}
}

iiopxx_mu_state::iiopxx_mu_state(pres_c_1 *mu_pres,
				 mu_state_op mu_op,
				 int mu_assumptions,
				 char *mu_which,
				 int swap)
	: mem_mu_state(mu_pres, mu_op, mu_assumptions, 2, 0, 8192/*XXX*/,
		       mu_which)
{
	should_swap = swap;
	principal_mark = -1;
}

iiopxx_mu_state::iiopxx_mu_state(const iiopxx_mu_state &must)
	: mem_mu_state(must)
{
	should_swap = must.should_swap;
	principal_mark = -1;
}

mu_state *iiopxx_mu_state::another(mu_state_op mu_op)
{
	return new iiopxx_mu_state(pres, mu_op, assumptions, get_which_stub(),
				   should_swap);
}

mu_state *iiopxx_mu_state::clone()
{
	return new iiopxx_mu_state(*this);
}

iiopxx_target_mu_state::iiopxx_target_mu_state(pres_c_1 *mu_pres,
					       mu_state_op mu_op,
					       int mu_assumptions,
					       char *mu_which,
					       int swap)
	: target_mu_state(mu_pres, mu_op, mu_assumptions, mu_which)
{
	should_swap = swap;
}

iiopxx_target_mu_state::iiopxx_target_mu_state(const iiopxx_target_mu_state &must)
	: target_mu_state(must)
{
	should_swap = must.should_swap;
}

mu_state *iiopxx_target_mu_state::clone()
{
	return new iiopxx_target_mu_state(*this);
}

char *iiopxx_target_mu_state::get_encode_name() 
{
	if (should_swap == IIOPXX_NO_SWAP || !(op & MUST_DECODE))
		return "cdr";
	else
		return "cdr_swap";
}

char *iiopxx_target_mu_state::get_be_name()
{
    return "iiopxx";
}

iiopxx_client_mu_state::iiopxx_client_mu_state(pres_c_1 *mu_pres,
					       mu_state_op mu_op,
					       int mu_assumptions,
					       char *mu_which,
					       int swap)
	: client_mu_state(mu_pres, mu_op, mu_assumptions, mu_which)
{
	should_swap = swap;
}

iiopxx_client_mu_state::iiopxx_client_mu_state(const iiopxx_client_mu_state &must)
	: client_mu_state(must)
{
	should_swap = must.should_swap;
}

mu_state *iiopxx_client_mu_state::clone()
{
	return new iiopxx_client_mu_state(*this);
}

char *iiopxx_client_mu_state::get_encode_name() 
{
	if (should_swap == IIOPXX_NO_SWAP || !(op & MUST_DECODE))
		return "cdr";
	else
		return "cdr_swap";
}

char *iiopxx_client_mu_state::get_be_name()
{
    return "iiopxx";
}

char *iiopxx_mu_state::get_encode_name() 
{
	if (should_swap == IIOPXX_NO_SWAP || now_packing || !(op & MUST_DECODE))
		return "cdr";
	else
		return "cdr_swap";
}

char *iiopxx_mu_state::get_be_name()
{
   return "iiopxx";
}

void
w_header_includes(pres_c_1 * /*p*/)
{
  w_printf("#include <flick/link/iiopxx.h>\n");
  w_printf("#include <flick/encode/cdr.h>\n");
}

/* End of file. */
