/*
 * Copyright (c) 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <assert.h>
#include <mom/c/libcast.h>
#include <mom/c/pbe.hh>

/*
 * This method handles out-of-line or out-of-band (pointer) data --- pointers
 * that are present because the data is not marshaled or unmarshaled to/from
 * the standard data stream, but passed via some other method (e.g., via the
 * kernel in Mach messages).
 *
 * This is similar to MAPPING_OPTIONAL_POINTER, in that the value of the
 * pointer itself conveys message data.  However, there is no actual data in
 * the message since it was passed via some other method.  It may be the case
 * that some marker or pointer exists in the message, which allows us to find
 * the real data, or it may be implicit by the simple fact that it was
 * transmitted out-of-line.
 *
 * In terms of MINT, this means that whatever we have for an itype here most
 * likely represents the real data being transmitted out-of-band.  We have no
 * direct need for it here, so we ignore it.
 *
 * THIS ROUTINE SHOULD BE SPECIALIZED to handle the specific requirements and
 * semantics of the transport mechanisms.
 */
void mu_state::mu_mapping_outofband(
	cast_expr pexpr,
	cast_type ctype,
	mint_ref /*itype*/,
	pres_c_mapping_outofband *obmap)
{
	/*
	 * If the CAST type of our pointer is a named type, locate the actual
	 * pointer type.  (Note: The MIG FE/PG makes these kinds of direct
	 * associations between named C types and MAPPING_POINTER nodes.  Other
	 * PG's generally indirect through a MAPPING_STUB, and the named type
	 * is derefenced to a pointer type there.)
	 */
	ctype = cast_find_typedef_type(&(pres->cast), ctype);
	if (!ctype)
		panic("In `mu_state::mu_mapping_pointer', "
		      "can't locate `typedef' for a named type.");
	
	assert(ctype->kind == CAST_TYPE_POINTER);
	cast_type to_ctype = ctype->cast_type_u_u.pointer_type.target;
	
	/* Allocate space */
	if ((op & MUST_ALLOCATE) &&
	    (ctype->kind == CAST_TYPE_POINTER))
		mu_pointer_alloc(pexpr, to_ctype, obmap->arglist_name);
	
	/*
	 * We don't really have a good ``default'' thing to do.  This should be
	 * specialized in the back end if anything more than magic is to occur.
	 */
	static int warned = 0;
	if (!warned++)
		warn("This back end doesn't support out-of-band data.");
	
	/* Deallocate space */
	if ((op & MUST_DEALLOCATE) &&
	    (ctype->kind == CAST_TYPE_POINTER))
		mu_pointer_free(pexpr, to_ctype, obmap->arglist_name);
}

/* End of file. */

