/*
 * Copyright (c) 1998, 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <assert.h>
#include <limits.h>

#include <mom/compiler.h>
#include <mom/c/libcast.h>
#include <mom/c/pbe.hh>

#define   min(a,b)        (a < b ? a : b)

/*
  A 'span' is considered to be a certain amount of data inside of a
  message.  We use the idea of a span to check for the availability
  of that message data for the unmarshaling code.  Therefore, we are able
  ensure that a bogus message doesn't cause the unmarshaling
  code to do bad things.  The implementation of spans is done by using
  a mu_span_node to represent some amount of the message.  The
  span node also corresponds to a check for that part of the span.
  Whenever there is some variably sized section of the message a child
  is added in order to represent it and the check that needs to be
  performed.  However, if there is a statically sized message section
  followed by a variably sized section followed by another statically
  sized message we will generate extra check calls.  The solution
  is simply to cause the previous checks to also check for anything
  following it.  Here's an example:

                        some rpc message
  ----------------------------------------------------------
  |                 |    variable length   |               |
  ----------------------------------------------------------
  ^                 ^
  Check span        check
  
  By making the variably sized span absorb the length of the span
  after it we only need to do two checks.  In the case of unions
  we need to add each case as a child of the current node.  Then
  we can simply take the minimum of all the children and absorb
  them into the parent, and then the siblings before it can also
  absorb the length.
  
  Note: The reason that we run a second pass over the tree is so
  that it can be constructed in different ways and still work.  If
  all the children of a node are connected to it directly rather
  than indirectly through extra parent nodes we can generate different
  arrangements of when we check a span.  For example, if we had two
  union's in a row we can generate two different trees:
  
       4               /-4
      /                |
     4-8               |-8
    / \                |
   /   16              |-16
  2                  2-+
   \   2               |-2
    \ /                |
     2-4               |-4
      \                |
       4               \-4
       
  The tree on the right is how it would currently be built and
  it requires the second pass since its hard to know when a parent
  is completely filled up and we can do a commit.  The tree on the left
  shows an alternative way where we could do a commit for the leaves
  during the processing of the union and thus we could avoid the second
  pass.  However, if you look closely the tree on the left has an extra
  check for two cases, whereas the tree on the right has only two.  They
  end up with the same number of checks but we do more in the formation
  on the left, we don't want that.

  XXX - This implementation works, however, it is flawed in that we
  don't separately represent each message section with a span node.
  Rather we use a single node and play games with the children's
  length expressions in order to get them to check spans that come
  after variable length spans.  This results in a less than optimal
  check_span count, which is why it needs to be fixed.

  It might also be possible to replace globbing with spans since the
  spans know exactly how much data there will be at a level.  A simple
  way to do this might be to just suppress any output from globbing code
  and simply have the span nodes output "flick_(link)_make_span(size,error)"
  macros.
*/

mu_span_node::mu_span_node(char *be_name)
{
	span_size = 0;
	span_start = 0;
	force_call = 0;
	make_call = 1;
	min_child_size = UINT_MAX;
	child_count = 0;
	/* Get an cast_expr to hold our span size */
	size_expr = cast_new_expr_lit_int(0,0);
	call_name = cast_new_expr_name(flick_asprintf("flick_%s_check_span",
						      be_name));
	macro_expr = cast_new_expr_call_2(call_name,
					  size_expr,
					  0);
	call_stmt = cast_new_stmt_expr(macro_expr);
	children = 0;
	stmt_added = 0;
	/* mod_expr defaults to size_expr, however, its possible
	   to change mod_expr to use add different expression */
	mod_expr = size_expr;
	abort_block = 0;
	conditional = 0;
}

mu_span_node::mu_span_node(cast_expr cex_name)
{
	span_size = 0;
	span_start = 0;
	force_call = 0;
	make_call = 1;
	min_child_size = UINT_MAX;
	child_count = 0;
	size_expr = cast_new_expr_lit_int(0,0);
	call_name = cex_name;
	macro_expr = cast_new_expr_call_2(call_name,
					  size_expr,
					  0);
	call_stmt = cast_new_stmt_expr(macro_expr);
	children = 0;
	stmt_added = 0;
	mod_expr = size_expr;
	abort_block = 0;
	conditional = 0;
}

mu_span_node::~mu_span_node()
{
}

mu_span_node *mu_span_node::add_child()
{
	mu_span_node *retval = new mu_span_node(call_name);
	
	child_count++;
	children = (mu_span_node **)mustrealloc(children,
						child_count *
						sizeof( mu_span_node * ));
	children[child_count - 1] = retval;
	return retval;
}

void mu_span_node::commit_child(mu_span_node *child)
{
	/* We need to remember where our parent started so that
	   we can check for any data that comes after us. */
	child->span_start = span_size;
	child->span_start_expr = mod_expr;
	/* Make the child absorb the minimum size of its children */
	child->collapse();
	/* If it hasn't been mod'ed then we just look at it's span_size */
	if( (child->mod_expr == child->size_expr) )
		min_child_size = min(min_child_size,
				     child->span_size);
	else {
		/* It has a special modifier expression so we need to
		   set the min_child_size to 0 since we don't really
		   have enough information to make a guess.  Of course
		   if the modifier expression is made up completely of
		   constant values then we can just add it to the parent.
		   This could also be helped by a cast_eval function
		   to see if the number is actually computable at compile
		   time so we can set a better min_child_size than 0. */
		if( !child->force_call &&
		    !this->conditional &&
		    cast_expr_const(child->mod_expr) ) {
			add_mod(cast_new_binary_expr(CAST_BINARY_ADD,
						     child->mod_expr,
						     0));
			child->make_call = 0;
		}
		min_child_size = 0;
	}
}

void mu_span_node::grow(int size)
{
	span_size += size;
}

void mu_span_node::shrink(int size)
{
	span_size -= size;
}

void mu_span_node::collapse()
{
	int lpc;

	if( child_count ) {
		for( lpc = 0; lpc < child_count; lpc++ ) {
			/* We stop the child from making a call if it
			   is the same as min_child_size since it was
			   completely absorbed by the parent.  If we
			   were to shrink all children by min_child_size
			   it wouldn't work since the pointer into the
			   message buffer hasn't been moved yet so the
			   check would be shorter than it should be. */
			if( (children[lpc]->mod_expr ==
			     children[lpc]->size_expr) &&
			    !children[lpc]->force_call &&
			    (children[lpc]->span_size == min_child_size) )
				children[lpc]->make_call = 0;
		}
		grow(min_child_size);
	}
}

void mu_span_node::commit()
{
	int lpc;

	/* Commit all the children first */
	for( lpc = 0; lpc < child_count; lpc++ ) {
		/* If the child has a modifier expression we need to
		   add another modifier in order to check for the rest
		   of our (the parent's) span.  Unfortunately, due to
		   the structure of span's we need to abuse the compiler
		   in order to get the right size value out. If this
		   was done in a way similar to what was mentioned above
		   we would still abuse the compiler, but it wouldn't be
		   as bad. */
		if( children[lpc]->make_call &&
		    (children[lpc]->mod_expr != children[lpc]->size_expr) &&
		    (span_size != children[lpc]->span_start) ) {
			/* We modify the child to check the rest of our span.
			   This is done by subtracting our current mod_expr
			   from our mod_expr from when the child is created.
			   The diff creates the right number, however, it
			   isn't pretty. */
			children[lpc]->add_mod(
				cast_new_binary_expr(
					CAST_BINARY_ADD,
					cast_new_binary_expr(
						CAST_BINARY_ADD,
						cast_new_binary_expr(
							CAST_BINARY_SUB,
							mod_expr,
							children[lpc]->
							span_start_expr),
						cast_new_expr_lit_int(
							span_size -
							children[lpc]->
							span_start, 0)
						),
					0));
		}
		children[lpc]->commit();
	}
	/* Fix size_expr to be span_size rather than 0 */
	size_expr->cast_expr_u_u.lit_prim.u.cast_lit_prim_u_u.i = span_size;
	/* If we don't need to make a call than
	   simply set it to a CAST_STMT_EMPTY. */
	if( !make_call ) {
		call_stmt->kind = CAST_STMT_EMPTY;
		if( abort_block ) {
			/* Its possible for the span to have used the last
			   abort label, which isn't needed anymore because
			   the call is now gone.  If this is the case we
			   need to decrement the use count for the label so
			   that it can be removed by the abort code. */
			abort_block->drop_label(
				macro_expr->cast_expr_u_u.call.params.
				cast_expr_array_val[1]->cast_expr_u_u.
				name.cast_scoped_name_val[0].name);
		}
	}
	else
		macro_expr->cast_expr_u_u.call.params.cast_expr_array_val[0] =
			mod_expr;
}

void mu_span_node::add_call(cast_stmt block, mu_abort_block *a_node)
{
	/* Only add if necessary */
	if( block && !stmt_added ) {
		abort_block = a_node;
		macro_expr->cast_expr_u_u.call.params.cast_expr_array_val[1] =
			cast_new_expr_name(abort_block->use_current_label());
		cast_block_add_stmt(&block->cast_stmt_u_u.block,call_stmt);
		stmt_added = 1;
	}
}

void mu_span_node::add_mod(cast_expr cex)
{
	switch( cex->kind ) {
	case CAST_EXPR_UNARY:
		cex->cast_expr_u_u.unary.expr = mod_expr;
		break;
	case CAST_EXPR_BINARY:
		cex->cast_expr_u_u.binary.expr[1] = mod_expr;
		break;
	default:
		assert( 0 );
		break;
	}
	mod_expr = cex;
}

void mu_span_node::adjust_child_starts()
{
	int lpc;
	
	for( lpc = 0; lpc < this->child_count; lpc++ ) {
		this->children[lpc]->span_start_expr = this->mod_expr;
	}
}
