/*
 * Copyright (c) 1999 The University of Utah and
 * the Computer Systems Laboratory at the University of Utah (CSL).
 *
 * This file is part of Flick, the Flexible IDL Compiler Kit.
 *
 * Flick is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Flick is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Flick; see the file COPYING.  If not, write to
 * the Free Software Foundation, 59 Temple Place #330, Boston, MA 02111, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <mom/c/scml.hh>

static char* scml_std_include_dirs_array[]
	= {
		".",
		SCML_INSTALL_DIR,	/* Set by Makefile. */
		0			/* Terminating null string. */
	};

char **scml_std_include_dirs()
{
	return scml_std_include_dirs_array;
}

void scml_alert(struct scml_stream_pos *ssp, int alert_flags, char *str, ...)
{
	char *alert_type = 0, *alert_class = 0;
	va_list args;
	
	va_start( args, str );
	switch( alert_flags & SAF_ALERT_CLASS ) {
	case SAF_INTERNAL:
		alert_class = "Internal";
		break;
	case SAF_GENERAL:
		alert_class = "General";
		break;
	case SAF_IO:
		alert_class = "I/O";
		break;
	case SAF_LEXICAL:
		alert_class = "Lexical";
		break;
	case SAF_PARSE:
		alert_class = "Parse";
		break;
	case SAF_TYPE:
		alert_class = "Type";
		break;
	case SAF_RUNTIME:
		alert_class = "Runtime";
		break;
	default:
		panic("Invalid alert class 0x%x passed to scml_alert()",
		      alert_flags & SAF_ALERT_CLASS);
		break;
	}
	switch( alert_flags & SAF_ALERT_TYPE ) {
	case SAF_WARNING:
		alert_type = "Warning";
		break;
	case SAF_ERROR:
		alert_type = "Error";
		break;
	default:
		panic("Invalid alert type 0x%x passed to scml_alert()",
		      alert_flags & SAF_ALERT_TYPE);
		break;
	}
	if( ssp ) {
		fprintf(stderr, "%s %s %s[",
			alert_class, alert_type,
			ssp->get_stream()->get_desc());
		if( ssp->get_last_row() == -1 ) {
			fprintf(stderr, "%d:%d] - ",
				ssp->get_row(), ssp->get_column());
		} else {
			fprintf(stderr, "%d:%d - %d:%d] - ",
				ssp->get_last_row(), ssp->get_last_column(),
				ssp->get_row(), ssp->get_column());
		}
	}
	else {
		fprintf(stderr, "%s %s - ", alert_class, alert_type);
	}
	vfprintf(stderr, str, args);
	fprintf(stderr, "\n");
	va_end( args );
}

int scml_hash_name(char *name, int table_size)
{
	int h;
	
	for( h = 0; *name; name++ )
		h = (64 * h + tolower(*name)) % table_size;
	return( h );
}

scml_string::scml_string()
{
	this->components.val = 0;
	this->components.len = 0;
}

scml_string::~scml_string()
{
	free( this->components.val );
}

char *scml_string::tag_ref()
{
	return( ptr_to_tag_ref("struct scml_string *", this) );
}

struct scml_string *scml_string::ptr(char *ref)
{
	return( (struct scml_string *)
		tag_ref_to_ptr("struct scml_string *", ref) );
}

char *scml_string::tag_string(tag_item *ti)
{
	char *retval = 0;
	
	switch( ti->data.kind ) {
	case TAG_STRING:
		retval = ti->data.tag_data_u.str;
		break;
	case TAG_REF:
		retval = scml_string::ptr(ti->data.tag_data_u.ref)->
			make_chars();
		break;
	default:
		break;
	}
	return( retval );
}

tag_data *scml_string::add_component()
{
	tag_data *retval;
	int i;
	
	i = this->components.len++;
	this->components.val = (tag_data *)mustrealloc(this->components.val,
						       sizeof( tag_data ) *
						       this->components.len);
	this->components.val[i].kind = TAG_NONE;
	retval = &this->components.val[i];
	return( retval );
}

void scml_string::concat(struct scml_string *ss)
{
	int start, lpc;
	
	start = this->components.len;
	this->components.len += ss->components.len;
	this->components.val = (tag_data *)mustrealloc(this->components.val,
						       sizeof( tag_data ) *
						       this->components.len);
	for( lpc = 0; lpc < ss->components.len; lpc++ ) {
		this->components.val[start + lpc] = ss->components.val[lpc];
	}
}

int scml_string::cmp(struct scml_string *ss)
{
	int my_pos = 0, ss_pos = 0, done = 0;
	char *my_str, *ss_str;
	int retval = 0;
	
	my_str = "";
	ss_str = "";
	while( !done ) {
		while( !(*my_str) && (my_pos < this->components.len) &&
		       (this->components.val[my_pos].kind == TAG_STRING) ) {
			my_str = this->components.val[my_pos].tag_data_u.str;
			my_pos++;
		}
		while( !(*ss_str) && (ss_pos < ss->components.len) &&
		       (ss->components.val[ss_pos].kind == TAG_STRING) ) {
			ss_str = ss->components.val[ss_pos].tag_data_u.str;
			ss_pos++;
		}
		while( *my_str && *ss_str && (*my_str == *ss_str) ) {
			my_str++;
			ss_str++;
		}
		if( *my_str < *ss_str )
			retval = -1;
		else if( *my_str > *ss_str )
			retval = 1;
		if( retval || (!*my_str && !*ss_str) ) {
			done = 1;
		}
	}
	return( retval );
}

char *scml_string::make_chars()
{
	char *retval = 0, *curr;
	int lpc, poss = 1, str_size = 0;
	
	for( lpc = 0; poss && (lpc < this->components.len); lpc++ ) {
		switch( this->components.val[lpc].kind ) {
		case TAG_BOOL:
			str_size += this->components.val[lpc].
				tag_data_u.b ? 4 : 5;
			break;
		case TAG_STRING:
			str_size += strlen(this->components.val[lpc].
					    tag_data_u.str);
			break;
		case TAG_INTEGER: {
			int i;
			
			i = this->components.val[lpc].tag_data_u.i;
			str_size++;
			while( i ) {
				i /= 10;
				str_size++;
			}
			break;
		}
		case TAG_FLOAT: {
			char f[30];
			
			sprintf(f, "%f", this->components.val[lpc].
				tag_data_u.f);
			str_size += strlen(f);
			break;
		}
		default:
			poss = 0;
			break;
		}
	}
	if( poss ) {
		retval = (char *)mustmalloc(str_size + 1);
		curr = retval;
		for( lpc = 0; lpc < this->components.len; lpc++ ) {
			switch( this->components.val[lpc].kind ) {
			case TAG_BOOL:
				strcpy(curr,
				       this->components.val[lpc].
				       tag_data_u.b ? "true" : "false");
				break;
			case TAG_STRING:
				strcpy(curr,
				       this->components.val[lpc].
				       tag_data_u.str);
				break;
			case TAG_INTEGER:
				sprintf(curr, "%d", this->components.
					val[lpc].tag_data_u.i);
				break;
			case TAG_FLOAT:
				sprintf(curr, "%f", this->components.
					val[lpc].tag_data_u.f);
				break;
			default:
				break;
			}
			curr += strlen(curr);
		}
	}
	return( retval );
}

void scml_string::print()
{
	union tag_data_u data;
	int lpc;
	
	for( lpc = 0; lpc < this->components.len; lpc++ ) {
		data = get_tag_data(&this->components.val[lpc], 0);
		print_tag_data(0, this->components.val[lpc].kind, data);
	}
}
